# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import

import unittest
import os
import sys

from ..pypi import _extract_html
from ..unpack import unpack_html


class ExtractHtmlTest(unittest.TestCase):

    def test_extract_html(self):
        path = os.path.join(os.path.dirname(__file__),
                            './fake_simple_html.txt')
        with open(path) as f:
            names = _extract_html(f.read())
        self.assertListEqual(names, 'a b c d e f g'.split())


class UnpackHtmlTest(unittest.TestCase):

    @unittest.skipIf(sys.version_info[0] != 3, 'Not python 3.x')
    def test_py3_unpack_html(self):
        data = bytes('cde', 'utf-8')
        self.assertEqual(unpack_html(data), data.decode('utf-8'))

    @unittest.skipIf(sys.version_info[0] != 2, 'Not python 2.x')
    def test_py2_unpack_html(self):
        data = 'abc'
        self.assertEqual(unpack_html(data), data.decode('utf-8'))



# -*- coding: utf-8 -*-

"""pypr - Python project requirements tool."""

from __future__ import print_function, division, absolute_import

from .utils import PY32

if not PY32:
    from gevent import monkey
    monkey.patch_all()



# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import


from ..utils import PY32


if PY32:
    from .thread_extractor import ThreadExtractor as Extractor
else:
    from .gevent_extractor import GeventExtractor as Extractor



# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import

import concurrent.futures

from .extractor import BaseExtractor

from ..log import logger


class ThreadExtractor(BaseExtractor):
    """Extractor use thread pool execute tasks.

    Can be used to extract /simple/<pkg_name> or /pypi/<pkg_name>/json.

    FIXME: can not deliver SIG_INT to threads in Python 2.
    """

    def __init__(self, names, max_workers=None):
        super(self.__class__, self).__init__(names, max_workers)
        self._futures = dict()

    def extract(self, job):
        """Extract url by package name."""
        with concurrent.futures.ThreadPoolExecutor(
                max_workers=self._max_workers) as executor:
            for name in self._names:
                self._futures[executor.submit(job, name)] = name

    def wait_complete(self):
        """Wait for futures complete done."""
        for future in concurrent.futures.as_completed(self._futures.keys()):
            try:
                error = future.exception()
            except concurrent.futures.CancelledError:
                break
            name = self._futures[future]
            if error is not None:
                err_msg = 'Extracting "{0}", got: {1}'.format(name, error)
                logger.error(err_msg)

    def shutdown(self):
        for future in self._futures:
            future.cancel()



# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import

import sys

import greenlet
from gevent.pool import Pool

from .extractor import BaseExtractor
from ..log import logger
from ..utils import Color


class GeventExtractor(BaseExtractor):

    def __init__(self, names, max_workers=222):
        super(self.__class__, self).__init__(names, max_workers)
        self._pool = Pool(self._max_workers)
        self._exited_greenlets = 0

    def extract(self, job):
        job = self._job_wrapper(job)
        for name in self._names:
            if self._pool.full():
                self._pool.wait_available()
            self._pool.spawn(job, name)

    def _job_wrapper(self, job):
        def _job(name):
            result = None
            try:
                result = job(name)
            except greenlet.GreenletExit:
                self._exited_greenlets += 1
            except Exception:
                e = sys.exc_info()[1]
                logger.error('Extracting "{0}", got: {1}'.format(name, e))
            return result
        return _job

    def wait_complete(self):
        self._pool.join()

    def shutdown(self):
        self._pool.kill(block=True)

    def final(self):
        count = self._exited_greenlets
        if count != 0:
            print(
                Color.YELLOW(
                    '** {0} running job exited.'.format(count)
                )
            )



# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import

import json
import collections
try:  # py2
    from urllib2 import urlopen, Request
    from HTMLParser import HTMLParser
except ImportError:  # py3
    from urllib.request import urlopen, Request
    from html.parser import HTMLParser


from .db import database
from .unpack import top_level, unpack_html
from .log import logger
from .utils import Color, compare_version, cmp_to_key
from .extractor import Extractor


PYPI_URL = 'https://pypi.python.org/'
PKG_URL = 'https://pypi.python.org/pypi/{0}'
PKGS_URL = 'https://pypi.python.org/simple/'
PKG_INFO_URL = 'https://pypi.python.org/pypi/{0}/json'
ACCEPTABLE_EXT = ('.whl', '.egg', '.tar.gz', '.tar.bz2', '.zip')


def search_names(names, installed_pkgs):
    """Search package information by names(`import XXX`).
    """
    results = collections.defaultdict(list)
    not_found = list()
    for name in names:
        logger.info('Searching package name for "{0}" ...'.format(name))
        # If exists in local environment, do not check on the PyPI.
        if name in installed_pkgs:
            results[name].append(list(installed_pkgs[name]) + ['local'])
        # Check information on the PyPI.
        else:
            rows = None
            with database() as db:
                rows = db.query_all(name)
            if rows:
                for row in rows:
                    version = extract_pkg_version(row.package)
                    results[name].append((row.package, version, 'PyPI'))
            else:
                not_found.append(name)
    return results, not_found


def check_latest_version(package):
    """Check package latest version in PyPI."""
    version = extract_pkg_version(package)
    return version


def update_db():
    """Update database."""
    print(Color.BLUE('Starting update database (this will take a while)...'))
    logger.info('Crawling "{0}" ...'.format(PKGS_URL))
    try:
        data = download(PKGS_URL)
    except Exception:
        logger.error("Fetch all packages got: ", exc_info=True)
        print(Color.RED('Operation abort ...'))
        return

    logger.info('Extracting all packages ...')
    pkg_names = _extract_html(unpack_html(data))
    with database() as db:
        ignore_pkgs = db.query_package(None)
        pkg_names = list(set(pkg_names) - set(ignore_pkgs))
    extractor = Extractor(pkg_names)
    extractor.run(extract_pkg_info)
    print(Color.BLUE('Operation done!'))


def extract_pkg_info(pkg_name):
    """Extract package information from PyPI."""
    logger.info('Extracting information of package "{0}".'.format(pkg_name))
    data = _pkg_json_info(pkg_name)
    # Extracting names which can be imported.
    if not data or not data['urls']:
        logger.warning('Package "{0}" no longer available.'.format(pkg_name))
        return

    urls = [item['url'] for item in data['urls']
            if item['filename'].endswith(ACCEPTABLE_EXT)]
    # Has not satisfied compressed package.
    if not urls:
        logger.warning('Package "{0}" can not unpack.'.format(pkg_name))
        return
    url = urls[0]

    top_levels = top_level(url, download(url))
    # Maybe package is a project, not importable...
    if not top_levels:
        logger.warning(
            'Maybe package "{0}" is not importable.'.format(pkg_name))
        return

    # Insert into database.
    with database() as db:
        db.insert_package(pkg_name)
        package = db.query_package(pkg_name)
        for top in top_levels:
            top = top or pkg_name  # empty top_level.txt
            db.insert_name(top, package.id)


def extract_pkg_version(pkg_name):
    """Extract package latest version from PyPI."""
    data = _pkg_json_info(pkg_name)
    if not data or not data['releases'] or not data['urls']:
        return 'unknown'
    latest = data['info'].get('version', None)
    if latest is None:
        latest = sorted(data['releases'], key=cmp_to_key(compare_version))
        latest = latest[-1]
    return latest


def _pkg_json_info(pkg_name):
    data = download(PKG_INFO_URL.format(pkg_name))
    if not data:  # 404
        return None
    data = json.loads(data.decode('utf-8'))
    return data


# Fake headers, just in case.
_HEADERS = {
    'Accept': '*/*',
    'Accept-Encoding': 'gzip, deflate, sdch',
    'Accept-Language': 'en-US,en;q=0.8,zh-CN;q=0.6,zh;q=0.4',
    'User-Agent': ('Mozilla/5.0 (X11; Linux x86_64; rv:13.0) '
                   'AppleWebKit/537.36 (KHTML, like Gecko) '
                   'Chrome/44.0.2403.157 Safari/537.36'),
}


def download(url, headers=_HEADERS):
    """Download data from url."""
    f = None
    try:
        f = urlopen(Request(url, headers=headers))
        data = f.read()
    finally:
        if f:
            f.close()
    return data


def _extract_html(html):
    """Extract data from html."""
    names = list()

    class HrefParser(HTMLParser):
        def handle_starttag(self, tag, attrs):
            if tag == 'a':
                attrs = dict(attrs)
                if attrs.get('href', None):
                    names.append(attrs['href'])

    HrefParser().feed(html)
    return names



# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import

import multiprocessing

from ..utils import Color


class BaseExtractor(object):

    def __init__(self, names, max_workers=None):
        self._names = names
        self._max_workers = max_workers or (multiprocessing.cpu_count() * 4)

    def run(self, job):
        try:
            self.extract(job)
            self.wait_complete()
        except KeyboardInterrupt:
            print(Color.BLUE('** Shutting down ...'))
            self.shutdown()
        else:
            print(Color.BLUE('^.^ Extracting all packages done!'))
        finally:
            self.final()

    def extract(self, job):
        raise NotImplemented

    def wait_complete(self):
        raise NotImplemented

    def shutdown(self):
        raise NotImplemented

    def final(self):
        pass
